use adventofcode2023::utils::get_puzzle_bytes;
use std::str;

type CubeSet = (u32, u32, u32);
type Game = (u32, Vec<CubeSet>);

fn parse_number(line: &[u8]) -> u32 {
    line.iter()
        .filter(|b| b.is_ascii_digit())
        .fold(0, |acc, &b| acc * 10 + (b - b'0') as u32)
}

fn parse_cube_set(set_str: &[u8]) -> CubeSet {
    let cubes_list = set_str.split(|&b| b == b',');
    cubes_list.fold(
        (0, 0, 0),
        |(red_cubes, green_cubes, blue_cubes), cube_str| {
            let pos = cube_str
                .iter()
                .position(|&b| b.is_ascii_alphabetic())
                .unwrap();
            let (cube_count_str, cube_color) = cube_str.split_at(pos);
            let cube_count = parse_number(cube_count_str);
            match cube_color {
                b"red" => (red_cubes + cube_count, green_cubes, blue_cubes),
                b"green" => (red_cubes, green_cubes + cube_count, blue_cubes),
                b"blue" => (red_cubes, green_cubes, blue_cubes + cube_count),
                _ => panic!("unknown color: {}", str::from_utf8(cube_color).unwrap()),
            }
        },
    )
}

fn parse_game_sets(game_str: &[u8]) -> Vec<CubeSet> {
    game_str
        .split(|&c| c == b';')
        .map(parse_cube_set)
        .collect::<Vec<CubeSet>>()
}

fn parse_puzzle(puzzle: &[u8]) -> Vec<Game> {
    puzzle
        .split(|&c| c == b'\n')
        .map(|line| {
            let sep = line.iter().position(|&c| c == b':').unwrap();
            line.split_at(sep)
        })
        .map(|(game_id_str, game_str)| {
            let game_id = parse_number(game_id_str);
            let sets_of_cubes = parse_game_sets(&game_str[1..]);
            (game_id, sets_of_cubes)
        })
        .collect::<Vec<Game>>()
}

fn check_max_cubes(sets_of_cubes: &[CubeSet]) -> bool {
    const MAX_RED_CUBES: u32 = 12;
    const MAX_GREEN_CUBES: u32 = 13;
    const MAX_BLUE_CUBES: u32 = 14;
    for (red_cubes, green_cubes, blue_cubes) in sets_of_cubes {
        if red_cubes > &MAX_RED_CUBES
            || green_cubes > &MAX_GREEN_CUBES
            || blue_cubes > &MAX_BLUE_CUBES
        {
            return false;
        }
    }
    true
}

fn compute_impossible_games_sum(puzzle: &[u8]) -> u32 {
    parse_puzzle(puzzle).iter().fold(0, {
        |acc, (game_id, sets_of_cubes)| {
            if check_max_cubes(sets_of_cubes) {
                acc + game_id
            } else {
                acc
            }
        }
    })
}

fn count_minimal_cubes(sets_of_cubes: &[CubeSet]) -> (u32, u32, u32) {
    sets_of_cubes.iter().fold((0, 0, 0), {
        |(red_min, green_min, blue_min), (red_cubes, green_cubes, blue_cubes)| {
            (
                if red_cubes > &red_min {
                    *red_cubes
                } else {
                    red_min
                },
                if green_cubes > &green_min {
                    *green_cubes
                } else {
                    green_min
                },
                if blue_cubes > &blue_min {
                    *blue_cubes
                } else {
                    blue_min
                },
            )
        }
    })
}

fn compute_sum_of_power(puzzle: &[u8]) -> u32 {
    parse_puzzle(puzzle).iter().fold(0, {
        |acc, (_, sets_of_cubes)| {
            let (red_min, green_min, blue_min) = count_minimal_cubes(sets_of_cubes);
            acc + red_min * green_min * blue_min
        }
    })
}

fn main() {
    let puzzle = get_puzzle_bytes();
    let start_time = std::time::Instant::now();
    println!("--- Day 2: Cube Conundrum ---");
    println!("What is the sum of the IDs of those games?");
    println!("{}", compute_impossible_games_sum(&puzzle));
    println!("--- Part Two ---");
    println!("What is the sum of the power of these sets?");
    println!("{}", compute_sum_of_power(&puzzle));
    let elapsed_time = start_time.elapsed();
    println!("compute time: {:?}", elapsed_time);
}
