use std::cmp::min;
use std::fmt::Debug;

use adventofcode2023::utils::get_puzzle_bytes;
use atoi::atoi;

#[derive(Debug)]
struct PartNumber {
    number: u32,
    range: std::ops::Range<usize>,
    valid: bool,
}

type PartsLine = Vec<PartNumber>;

trait SymbolCheckable {
    fn is_symbol(&self) -> bool;
}

impl SymbolCheckable for u8 {
    fn is_symbol(&self) -> bool {
        *self != b'.' && !self.is_ascii_punctuation()
    }
}

fn parse_number(line: &[u8]) -> (u32, usize) {
    let number_len = line
        .iter()
        .position(|&c| !c.is_ascii_digit())
        .unwrap_or(line.len());
    let number = atoi::<u32>(&line[..number_len]).unwrap();
    (number, number_len)
}

fn parse_parts(puzzle: &[u8]) -> Vec<PartsLine> {
    puzzle
        .split(|&c| c == b'\n')
        .map(|line| {
            let mut part_numbers = Vec::new();
            let mut i = 0;
            while i < line.len() {
                let c = line.get(i).unwrap();
                if c.is_ascii_digit() {
                    let (number, number_len) = parse_number(&line[i..]);
                    let part_number = PartNumber {
                        number,
                        range: i.saturating_sub(1)..i + number_len + 1,
                        valid: false,
                    };
                    i = part_number.range.end;
                    part_numbers.push(part_number);
                } else {
                    i += 1;
                }
            }
            part_numbers
        })
        .collect::<Vec<PartsLine>>()
}

fn compute_part_numbers_sum(puzzle: &[u8]) -> u32 {
    let mut part_lines = parse_parts(puzzle);
    let line_range_max = part_lines.len() - 1;
    for (line_index, line) in puzzle.split(|&c| c == b'\n').enumerate() {
        for (i, c) in line.iter().enumerate() {
            if c.is_symbol() {
                let line_range = line_index.saturating_sub(1)..=min(line_index + 1, line_range_max);
                part_lines[line_range]
                    .iter_mut()
                    .flat_map(|v| v.iter_mut())
                    .filter(|part_number| !part_number.valid)
                    .for_each(|part_number| part_number.valid = part_number.range.contains(&i));
            }
        }
    }
    part_lines
        .iter()
        .flat_map(|v| v.iter())
        .filter(|part_number| part_number.valid)
        .fold(0, |acc, part_number| acc + part_number.number)
}

fn compute_gear_ratios_sum(puzzle: &[u8]) -> u32 {
    let part_lines = parse_parts(puzzle);
    let line_range_max = part_lines.len() - 1;
    let mut gear_ratios_sum = 0;
    for (line_index, line) in puzzle.split(|&c| c == b'\n').enumerate() {
        for (i, c) in line.iter().enumerate() {
            if *c == b'*' {
                let line_range = line_index.saturating_sub(1)..=min(line_index + 1, line_range_max);
                let gear_parts = part_lines[line_range]
                    .iter()
                    .flat_map(|v| v.iter())
                    .filter_map(|part_number| match part_number.range.contains(&i) {
                        true => Some(part_number.number),
                        false => None,
                    })
                    .collect::<Vec<u32>>();
                if gear_parts.len() == 2 {
                    gear_ratios_sum += gear_parts[0] * gear_parts[1];
                }
            }
        }
    }
    gear_ratios_sum
}

fn main() {
    let puzzle = get_puzzle_bytes();
    let start_time = std::time::Instant::now();
    println!("--- Day 3: Gear Ratios ---");
    println!("What is the sum of all of the part numbers in the engine schematic?");
    println!("{}", compute_part_numbers_sum(&puzzle));
    println!("--- Part Two ---");
    println!("What is the sum of all of the gear ratios in your engine schematic?");
    println!("{}", compute_gear_ratios_sum(&puzzle));
    let elapsed_time = start_time.elapsed();
    println!("compute time: {:?}", elapsed_time);
}
