use adventofcode2023::utils::get_puzzle_bytes;
use atoi::atoi;
use std::array;
use std::cmp;
use std::iter;
use std::ops::Range;

const SECTIONS: usize = 7;

type Map = Vec<(Range<u64>, u64)>;

fn parse_maps(puzzle: &[u8]) -> Vec<Map> {
    let mut lines = puzzle.split(|&c| c == b'\n').skip(2);
    (0..SECTIONS)
        .map(|_| {
            let mut map = (&mut lines)
                .skip(1)
                .take_while(|line| !line.is_empty())
                .map(|line| {
                    let mut iter = line.splitn(3, |b| b == &b' ').map(|b| atoi(b).unwrap());
                    let values: [_; 3] = array::from_fn(|_| iter.next().unwrap());
                    (values[1]..values[1] + values[2], values[0])
                })
                .collect::<Map>();
            map.sort_by_key(|(range, _)| range.start);
            map
        })
        .collect()
}

fn parse_seeds(puzzle: &[u8]) -> Vec<u64> {
    puzzle[0..puzzle.iter().position(|b| b == &b'\n').unwrap()]
        .split(|b| b == &b' ')
        .skip(1)
        .map(|b| atoi(b).unwrap())
        .collect()
}

fn compute_lowest_location_number(puzzle: &[u8]) -> u64 {
    let maps = parse_maps(puzzle);
    let seeds = parse_seeds(puzzle);
    seeds
        .iter()
        .map(|&seed| {
            maps.iter().fold(seed, |seed, map| {
                map.iter()
                    .find_map(|(range, to)| match range.contains(&seed) {
                        true => Some(seed + to - range.start),
                        false => None,
                    })
                    .unwrap_or(seed)
            })
        })
        .min()
        .unwrap() as u64
}

fn parse_seed_ranges(puzzle: &[u8]) -> Vec<Range<u64>> {
    let mut range_iter = puzzle[0..puzzle.iter().position(|b| b == &b'\n').unwrap()]
        .split(|b| b == &b' ')
        .skip(1)
        .map(|b| atoi::<u64>(b).unwrap());
    iter::from_fn(|| range_iter.next().zip(range_iter.next()))
        .map(|(start, len)| start..start + len)
        .collect()
}

fn overlap<T: Ord>(a: Range<T>, b: Range<T>) -> Option<Range<T>> {
    let start = cmp::max(a.start, b.start);
    let end = cmp::min(a.end, b.end);
    if start < end {
        Some(start..end)
    } else {
        None
    }
}

fn map_range_to_destination(range: &Range<u64>, map: &Map) -> Vec<Range<u64>> {
    let mut mapped_ranges = Vec::new();
    let remainder_range = map.iter().fold(range.clone(), |range, (map_range, to)| {
        match overlap(range.clone(), map_range.clone()) {
            Some(overlap_range) => {
                if overlap_range.start > range.start {
                    mapped_ranges.push(range.start..overlap_range.start);
                }
                mapped_ranges.push(
                    overlap_range.start + to - map_range.start
                        ..overlap_range.end + to - map_range.start,
                );
                overlap_range.end..range.end
            }
            None => range,
        }
    });
    if !remainder_range.is_empty() {
        mapped_ranges.push(remainder_range);
    }
    mapped_ranges
}

fn compute_lowest_location_from_ranges(puzzle: &[u8]) -> u64 {
    let maps = parse_maps(puzzle);
    let seed_ranges = parse_seed_ranges(puzzle);
    seed_ranges
        .iter()
        .flat_map(|seed_range| {
            maps.iter().fold(vec![seed_range.clone()], |ranges, map| {
                ranges
                    .iter()
                    .flat_map(|range| map_range_to_destination(range, map))
                    .collect()
            })
        })
        .map(|range| range.start)
        .min()
        .unwrap()
}

fn main() {
    let puzzle = get_puzzle_bytes();
    let start_time = std::time::Instant::now();
    println!("--- Day 5: If You Give A Seed A Fertilizer ---");
    println!(
        "What is the lowest location number that corresponds to any of the initial seed numbers?"
    );
    println!("{}", compute_lowest_location_number(&puzzle));
    println!("--- Part Two ---");
    println!(
        "What is the lowest location number that corresponds to any of the initial seed numbers?"
    );
    println!("{}", compute_lowest_location_from_ranges(&puzzle));
    let elapsed_time = start_time.elapsed();
    println!("compute time: {:?}", elapsed_time);
}
